import 'package:global_state/src/flutter.dart' show protected;

import 'error.dart';

/// Required interface for event listeners that use [GSController].
///
/// Implemented by [GSElement].
abstract class Listener implements FastHashCode {
  @protected
  Set<Notifier> notifiers;

  /// `postponeEventHandling` advices the listener to postpone the event
  /// handling.
  @protected
  handleEvent(Notifier notifier, [bool postponeEventHandling = false]);
}

enum NotifierStatus {
  subscribing,
  notifying,
  defunct,
}

/// Required interface for notifiers that use [GSController].
///
/// Implemented by [GSNotifier], [GSValue] and [GSMap].
abstract class Notifier implements FastHashCode {
  @protected
  Set<Listener> listeners;

  NotifierStatus status;

  /// Notifies the controller about a read event.
  notifyRead();

  /// Notifies the listeners about a change event.
  notifyChange();
}

/// Class that subscribes [Listener] to [Notifier] instances.
class GSController {
  static Listener _activeListener;

  /// The object that is currently listening to [Notifier] instances.
  static Listener get activeListener => _activeListener;

  static bool get isListening => _activeListener != null;

  static startListening(Listener listener) {
    if (isListening) {
      // Triggers when there was an error during the build cycle.
      stopListening();
    }
    _activeListener = listener;
  }

  static void stopListening() {
    assert(_activeListener != null);
    assert(() {
      for (var notifier in _currentNotifiers) {
        if (notifier.status != NotifierStatus.subscribing) {
          return false;
        }
      }
      return true;
    }());
    _commitSubscriptions();
    assert(_currentNotifiers.isEmpty);
    _activeListener = null;
  }

  static Set<Notifier> _currentNotifiers = Set();

  static registerNotifier(Notifier notifier) {
    assert(isListening);
    assert(notifier.status != NotifierStatus.defunct);
    if (notifier.status != NotifierStatus.subscribing) {
      notifier.status = NotifierStatus.subscribing;
      notifier.listeners ??= Set();
      _currentNotifiers.add(notifier);
    }
  }

  static _commitSubscriptions() {
    final listener = _activeListener;
    final previousNotifiers = _activeListener.notifiers;
    final currentNotifiers = _currentNotifiers;
    assert(() {
      if (currentNotifiers.isNotEmpty &&
          (previousNotifiers == null || previousNotifiers.isEmpty)) {
        GSController._debugSubscribedListeners.add(_activeListener);
      }
      return true;
    }());
    if (previousNotifiers == null || previousNotifiers.isEmpty) {
      if (currentNotifiers.isNotEmpty) {
        for (var notifier in currentNotifiers) {
          notifier.listeners.add(listener);
          notifier.status = NotifierStatus.notifying;
        }
        listener.notifiers = currentNotifiers;
        _currentNotifiers = Set();
      }
    } else {
      int persistentNotifiers = 0;
      for (var oldNotifier in previousNotifiers) {
        if (isSubscribing(oldNotifier)) {
          persistentNotifiers++;
          oldNotifier.status = NotifierStatus.notifying;
        } else {
          // unsubscribe from non subscribing old notifiers.
          oldNotifier.listeners.remove(listener);
        }
      }
      if (persistentNotifiers != currentNotifiers.length) {
        // Subscribe listener to new notifiers.
        for (var notifier in currentNotifiers) {
          if (isSubscribing(notifier)) {
            notifier.listeners.add(listener);
            notifier.status = NotifierStatus.notifying;
          }
        }
      }
      listener.notifiers = currentNotifiers;
      _currentNotifiers = previousNotifiers..clear();
    }
    assert(() {
      for (var notifier in currentNotifiers) {
        if (notifier.status != NotifierStatus.notifying) {
          return false;
        }
      }
      return true;
    }());
  }

  /// Flag that prevents notification events from triggering an assertion error
  /// when listening.
  static bool debugAllowNotificationsWhenListening = false;

  static bool isSubscribing(notifier) =>
      notifier.status == NotifierStatus.subscribing;

  static notifyChangeToListeners(Notifier notifier,
      [bool advicePostponing = false]) {
    assert(() {
      if (isSubscribing(notifier) &&
          !advicePostponing &&
          !debugAllowNotificationsWhenListening) {
        gsError('`$activeListener` is listening (a widget is building) '
            'while setting value of the [Notifier] $notifier. '
            '[Notifier] instances like [GSMap] cannot be modified '
            'within a build method.\n'
            'In GS widgets, initContext can be used to initialize values '
            'values of GS maps, lists or values.');
        return false;
      }
      return true;
    }());
    for (var listener in notifier.listeners) {
      listener.handleEvent(notifier, advicePostponing);
    }
  }

  static unsubscribeListener(Listener listener) {
    assert(!isListening);
    for (var notifier in listener.notifiers) {
      assert(notifier.listeners.contains(listener));
      notifier.listeners.remove(listener);
    }
    listener.notifiers.clear();
    assert(() {
      GSController._debugSubscribedListeners.remove(listener);
      return true;
    }());
  }

  static unsubscribeNotifier(Notifier notifier) {
    for (var listener in notifier.listeners) {
      listener.notifiers.remove(notifier);
    }
    notifier.listeners.clear();
  }

  static final _debugSubscribedListeners = Set<Listener>();

  static int get debugSubscribedListenersCount =>
      _debugSubscribedListeners.length;

  static debugReset() {
    for (var listener in _debugSubscribedListeners.toList()) {
      unsubscribeListener(listener);
    }
  }

  static bool debugContainsListener(Listener listener) =>
      _debugSubscribedListeners.contains(listener);
}

/// Class to retrieve a fast hashCode.
///
/// The inherited Dart implementation of hashCode is very slow, it takes 10
/// times longer than using a fixed int. This causes a big performance drop in
/// hash data structures like Set or Map. Implementations os [Notifier]
/// and [Listener] should include this mixin.
mixin FastHashCode {
  static int _lastId = 0;

  /// The value returned by [hashCode].
  ///
  /// Implementers of [FastHashCode] interface that do not include it as mixin
  /// should override [hashcode] to return this value.
  final hashId = _lastId++;

  int get hashCode => hashId;
}

/// Implementation of [Notifier] using [GSController].
///
/// Mixed in by [GSNotifier], base class of [GSValue] and [GSMap].
abstract class NotifierMixin implements Notifier {
  Listener get activeListener => GSController._activeListener;
  bool get controllerIsListening => GSController.isListening;

  @protected
  Set<Listener> listeners;

  NotifierStatus status = NotifierStatus.notifying;

  bool get hasListeners => listeners != null && listeners.isNotEmpty;

  void notifyRead() {
    assert(status != NotifierStatus.defunct);
    if (GSController.isListening) {
      GSController.registerNotifier(this);
    }
  }

  void notifyChange([bool postponeNotificationHandling = false]) {
    assert(status != NotifierStatus.defunct);
    if (listeners != null) {
      if (listeners.isNotEmpty) {
        GSController.notifyChangeToListeners(
            this, postponeNotificationHandling);
      } else if (postponeNotificationHandling) {
        GSController.notifyChangeToListeners(this, true);
      }
    }
  }

  void forgetListeners() {
    assert(status != NotifierStatus.defunct);
    if (hasListeners) {
      GSController.unsubscribeNotifier(this);
    }
  }

  /// Dispose can be invoked when this notifier is not going to be used again.
  void dispose() {
    assert(status == NotifierStatus.notifying);
    forgetListeners();
    assert(() {
      status = NotifierStatus.defunct;
      return true;
    }());
  }
}
